
// Compute the mandelbrot set on the GPU
//
// Each pixel computes their own z-sequence,
// then sets the color depending on the #iterations
//
// nvcc -arch=sm_35 --compiler-options -Wall mandelbrot_gpu.cu 
//      -Xlinker -framework,OpenGL,-framework,GLUT  -o obj/mandelbrot_gpu.o

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#if defined(__APPLE__) || defined(__MACOSX__)
#include <GLUT/glut.h>
#else
// On Linux, include the system's copy of glut.h, glext.h, and glx.h
#include <GL/glut.h>
#include <GL/glext.h>
#include <GL/glx.h>
#endif


// TODO: SCALE als param; eigenen pixel data type, CUDA error checks; threads(16,16)
// TODO: implement some interactions (scaling, translation, ribbon width, ...)

// make the following parameters
int img_size = 1024;					// default image size
#define SCALE 1.4



////////////////////////////////////////////////////////////////////////
//    Device side stuff
////////////////////////////////////////////////////////////////////////


// Class providing a complex numbers type
struct cuComplex
{
    float   r;                 // real
    float   i;                 // imaginary

    __device__                 // constructor
	cuComplex( float a, float b ) : r(a), i(b)  {}

    __device__                // |z|^2
	float magnitude2( void )
	{
        return r * r + i * i;
    }

    __device__                // z1 * z2
	cuComplex operator * (const cuComplex & a)
	{
        return cuComplex(r*a.r - i*a.i, i*a.r + r*a.i);
    }

    __device__               // z1 *= scalar
	void operator *= (const float s)
	{
		r *= s;
		i *= s;
    }

    __device__               // z1 + z2
	cuComplex operator + (const cuComplex & a)
	{
        return cuComplex(r+a.r, i+a.i);
    }
};



// Convert HSV color space to RGB
// input:  h must be in [0,360); s,v must be in [0,1]
// output: r,g,b (out) will be in [0,1]
// Warning: no parameter range checks are done!
__device__
void HSVtoRGB( float *r, float *g, float *b, float h, float s, float v )
{
	if ( fabsf(s) < 1E-5 )
	{
		// achromatic (grey)
		*r = *g = *b = v;
		return;
	}

	h /= 60.0f;			// h = sector in [0, 5]
	int i = floor( h );
	float f, p, q, t;
	f = h - i;			// factorial part of h
	p = v * ( 1 - s );
	q = v * ( 1 - s * f );
	t = v * ( 1 - s * ( 1 - f ) );
	switch( i )
	{
		case 0:
			*r = v; *g = t; *b = p;
			break;
		case 1:
			*r = q; *g = v; *b = p;
			break;
		case 2:
			*r = p; *g = v; *b = t;
			break;
		case 3:
			*r = p; *g = q; *b = v;
			break;
		case 4:
			*r = t; *g = p; *b = v;
			break;
		default:		// and case 5
			*r = v; *g = p; *b = q;
	}
}



__device__
// Returns zero if (x,y) is in M, #iterations otherwise
unsigned int isPointInMandelbrot( int x, int y, const int img_size )
{
    const float scale = SCALE;
	cuComplex c( static_cast<float>(x - img_size/2) / (img_size/2),
				 static_cast<float>(y - img_size/2) / (img_size/2) );
	c *= scale;
	cuComplex z( 0.0, 0.0 );

	for ( unsigned int i = 0; i < 200; i ++ )
	{
		z = z*z + c;
		if ( z.magnitude2() > 4.0f )
			return i;
	}

	return 0;
}




__global__
// Kernel for computing the mandelbrot image
// we don't need to check in the kernel whether the thread is outside the image,
// because threads layout and image size are made such that no thread can be out of range!
void mandelImage( unsigned char *bitmap, const int img_size )
{
    // map from blockIdx to pixel position
    int x = blockIdx.x * blockDim.x + threadIdx.x;
    int y = blockIdx.y * blockDim.y + threadIdx.y;
    int offset = x + y * (gridDim.x * blockDim.x);

    // now calculate the value at that position
    unsigned int iterations = isPointInMandelbrot( x, y, img_size );
	//printf("(%d,%d)(%d,%d): %d\n",
		   //blockIdx.x, blockIdx.y, threadIdx.x, threadIdx.y, iterations );

	float h, s, v;
	if ( iterations == 0 )
		h = 0.0,  s = 0.0,  v = 0.0;			// black
	else
	{
		const int ribbon_width = 20;
		h = (float)(iterations % ribbon_width) / ribbon_width * 360.0;		// hue
		s = 1.0;
		v = 1.0;
	}
	float r, g, b;
	HSVtoRGB( &r, &g, &b, h, s, v );

    bitmap[offset*4 + 0] = 255 * r;
    bitmap[offset*4 + 1] = 255 * g;
    bitmap[offset*4 + 2] = 255 * b;
    bitmap[offset*4 + 3] = 255;
}




////////////////////////////////////////////////////////////////////////
//    Host side stuff
////////////////////////////////////////////////////////////////////////



unsigned char * h_bitmap;			// image bitmap on host side
unsigned char * d_bitmap;			// image bitmap on device side


// key handler for GLUT
void key( unsigned char key, int x, int y )
{
	switch (key)
	{
		case 27:
			exit(0);
	}
}


void draw( void )
{
	glClearColor( 0.0, 0.0, 0.0, 1.0 );
	glClear( GL_COLOR_BUFFER_BIT );
	glDrawPixels( img_size, img_size, GL_RGBA, GL_UNSIGNED_BYTE, h_bitmap );
	glFlush();
}



int main( int argc, char **argv )
{
	if ( argc > 1 )
		// interpret first argument as image size
		img_size = atoi( argv[1] );
	img_size = (img_size / 32) * 32;			// round to multiple of 32
	if ( img_size == 0 )   img_size = 256;
	printf( "Image size: %d\n", img_size );
	
	const unsigned int bitmap_size = img_size * img_size * 4;
	h_bitmap = new unsigned char[bitmap_size];
    cudaMalloc( (void**) &d_bitmap, bitmap_size );

	// set threads layout;  we utilize the fact that img_size = k*32 (see above)
	// consequence: blocks.x * threads.x = img_size
	// -> we don't need to check in the kernel whether the thread is outside the image!
	dim3 threads( 16, 16 );
    dim3 blocks( img_size/threads.x, img_size/threads.y );
	printf("Num blocks = %dx%d\n", blocks.x, blocks.y );
    mandelImage<<< blocks,threads >>>( d_bitmap, img_size );

    cudaMemcpy( h_bitmap, d_bitmap, bitmap_size, cudaMemcpyDeviceToHost );

    cudaFree( d_bitmap );

	int c=1; char* dummy = "";		// dummy argument for glutInit
	glutInit( &c, &dummy );
	glutInitDisplayMode( GLUT_SINGLE | GLUT_RGBA );
	glutInitWindowSize( img_size, img_size );
	glutCreateWindow( "Mandelbrot on GPU" );
	glutKeyboardFunc(key);
	glutDisplayFunc(draw);
	glutMainLoop();

}

