
// Add long vectors using CUDA
// Uses a 1D grid layout; if you have huge vectors, you'll need to use vectoradd_2d !
//
// nvcc -arch=sm_35 --compiler-options -Wall vectoradd.cu -o vectoradd
//
// GZ, Feb 2013


#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#ifdef _WIN32
#define srand48(expr) srand(expr)
#define drand48() double(rand())/double(RAND_MAX)
#endif

// Add two vectors 'a' and 'b' of length 'n', put the result in c.
// Prerequisite: 'c'c must hold at least 'n' floats!

__global__
void addVectors( const float *a, const float *b, float *c, unsigned int n )
{
	unsigned int i = blockDim.x * blockIdx.x + threadIdx.x;
	if ( i < n )
		c[i] = a[i] + b[i];
}



// General CUDA error handler
// Prints error message, if something went wrong
void handleCUDAerror( int line )
{
	cudaError_t err = cudaGetLastError();
	if ( err == cudaSuccess )
		return;

	printf("CUDA error message: %s\n"
		   "line %d\n", cudaGetErrorString(err), line );

	int devID;
    cudaGetDevice( &devID );
    cudaDeviceProp props;
    cudaGetDeviceProperties( &props, devID );
    printf("CUDA device %d: \"%s\" with compute capability %d.%d\n",
           devID, props.name, props.major, props.minor);
}


// From the current GPU, return 
// 1. maximum number of threads per block 
// 2. maximum number of threads per block along x dimension
// 3. maximum number of blocks along x dimension
void getMaxProperties( unsigned int * maxNumThreadsPerBlock,
					   unsigned int * maxNumThreadsPerBlockX,
					   unsigned int * maxNumBlocksX                                        )
{
	int devID;
    cudaGetDevice( &devID );
    cudaDeviceProp props;
    cudaGetDeviceProperties( &props, devID );
	handleCUDAerror(__LINE__);

	*maxNumThreadsPerBlock = static_cast<unsigned int>( props.maxThreadsPerBlock );
	*maxNumThreadsPerBlockX = static_cast<unsigned int>( props.maxThreadsDim[0] );
	*maxNumBlocksX = static_cast<unsigned int>( props.maxGridSize[0] );
}



int main( int argc, char **argv )
{
	unsigned int vec_len = 10000;									// just a default
	if ( argc > 1 )
		// interpret first argument as length
		vec_len = atoi( argv[1] );
	printf( "Add two vectors of length %d\n", vec_len );

	// fill two vectors with random numbers
	size_t size = vec_len * sizeof(float);
	float * h_a = (float*) malloc( size );
	float * h_b = (float*) malloc( size );
	float * h_c = (float*) malloc( size );
	if ( h_a == NULL || h_b == NULL || h_c == NULL )
    {
        fprintf(stderr, "Failed to allocate host vectors!\n");
		exit(1);
	}

	srand48( time(NULL) );
    for ( unsigned int i = 0; i < vec_len; i++ )
    {
        h_a[i] = drand48();
        h_b[i] = drand48();
    }
	
	// transfer the arrays to the device
	float *d_a, *d_b, *d_c;
	cudaMalloc( (void **) &d_a, size );
	cudaMalloc( (void **) &d_b, size );
	cudaMalloc( (void **) &d_c, size );
	handleCUDAerror(__LINE__);

	cudaMemcpy( d_a, h_a, size, cudaMemcpyHostToDevice );
	cudaMemcpy( d_b, h_b, size, cudaMemcpyHostToDevice );
	handleCUDAerror(__LINE__);									// yes, we should do this after every memcpy ...

	// set layout of blocks and threads; we make a 1D threads/blocks layout here
	unsigned int threads_per_block = 256;						// any 2^k in [1,1024]
	unsigned int num_blocks = (vec_len + threads_per_block - 1) / threads_per_block;
	// this funny expression yields ceil( max_blocks_per_grid / threads_per_block ),
	// but needs only integer arithmetic
	printf( "Threads per block = %d\n", threads_per_block );
	printf( "Number of blocks  = %d\n", num_blocks );

	// check maximally allowed layout sizes
	unsigned int max_n_threads_per_block, max_n_threads_x, max_n_blocks_x;
	getMaxProperties( & max_n_threads_per_block, & max_n_threads_x, & max_n_blocks_x );
	if ( threads_per_block > max_n_threads_per_block || threads_per_block > max_n_threads_x )
	{
		fprintf(stderr,"Too many threads per block! Max = %d / %d\n", max_n_threads_per_block, max_n_threads_x );
		exit(2);
	}
	if ( num_blocks > max_n_blocks_x )
	{
		fprintf(stderr,"Too many blocks in grid row! Max = %d\n", max_n_blocks_x );
		exit(2);
	}

	// launch the vector addition kernel
	addVectors<<<num_blocks,threads_per_block>>>( d_a, d_b, d_c, vec_len );
	handleCUDAerror(__LINE__);

	// copy results back to the host (does cudaDeviceSynchronize implicitely)
	cudaMemcpy( h_c, d_c, size, cudaMemcpyDeviceToHost );
	handleCUDAerror(__LINE__);

	// verify the results (kind of a unit test)
	bool verify_ok = true;
	for ( unsigned int i = 0; i < vec_len; i++ )
	{
		if ( fabsf(h_a[i] + h_b[i] - h_c[i]) > 1e-5 )
		{
			fprintf(stderr, "Error at element %d: %f + %f != %f !\n",
					i, h_a[i], h_b[i], h_c[i] );
			verify_ok = false;
		}
	}

	// free all memory
	cudaFree( d_a );
	cudaFree( d_b );
	cudaFree( d_c );
	handleCUDAerror(__LINE__);

	free( h_a );
	free( h_b );
	free( h_c );

	cudaDeviceReset();
	if ( verify_ok )
		puts("Success.");
	return 0;
}

